###
# Copyright (c) 2005, Jeremiah Fincher
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import supybot.conf as conf
import supybot.registry as registry
from supybot.i18n import PluginInternationalization, internationalizeDocstring
_ = PluginInternationalization('Later')

def configure(advanced):
    # This will be called by setup.py to configure this module.  Advanced is
    # a bool that specifies whether the user identified themself as an advanced
    # user or not.  You should effect your configuration by manipulating the
    # registry as appropriate.
    from supybot.questions import expect, anything, something, yn
    conf.registerPlugin('Later', True)

Later = conf.registerPlugin('Later')
conf.registerGlobalValue(Later, 'maximum',
    registry.NonNegativeInteger(0, _("""Determines the maximum number of
    messages to be queued for a user.  If this value is 0, there is no maximum.
    """)))
conf.registerGlobalValue(Later, 'private',
    registry.Boolean(False, _("""Determines whether users will be notified in
    the first place in which they're seen, or in private.""")))
conf.registerGlobalValue(Later, 'tellOnJoin',
    registry.Boolean(False, _("""Determines whether users will be notified upon
    joining any channel the bot is in, or only upon sending a message.""")))
conf.registerGlobalValue(Later, 'messageExpiry',
    registry.NonNegativeInteger(30, _("""Determines the maximum number of
    days that a message will remain queued for a user. After this time elapses,
    the message will be deleted. If this value is 0, there is no maximum.""")))

conf.registerGroup(Later, 'format')
conf.registerGlobalValue(Later.format, 'senderHostname',
    registry.Boolean(False, _("""Determines whether senders' hostname will be
    shown in messages (instead of just the nick).""")))

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



#!/usr/bin/env python3

###
# Copyright (c) 2005, Ali Afshar
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

import os
import sys
import shutil
import string
import logging

import supybot

def error(s):
    sys.stderr.write('%s\n' % s)
    sys.exit(-1)

# We need to do this before we import conf.
if not os.path.exists('doc-conf'):
    os.mkdir('doc-conf')

registryFilename = os.path.join('doc-conf', 'doc.conf')
try:
    fd = open(registryFilename, 'w')
    fd.write("""
supybot.directories.data: doc-data
supybot.directories.conf: doc-conf
supybot.directories.log: doc-logs
supybot.log.stdout: False
supybot.log.level: DEBUG
supybot.log.format: %(levelname)s %(message)s
supybot.log.plugins.individualLogfiles: False
supybot.databases: sqlite3 anydbm cdb flat pickle
""")
    fd.close()
except EnvironmentError as e:
    error('Unable to open %s for writing.' % registryFilename)

import supybot.registry as registry
registry.open_registry(registryFilename)

import supybot.log as log
import supybot.conf as conf
conf.supybot.flush.setValue(False)

import textwrap

import supybot.utils as utils
import supybot.world as world
import supybot.plugin as plugin
import supybot.registry as registry

world.documenting = True

TITLE_TEMPLATE = 'Documentation for the $name plugin for Supybot'

class PluginDoc(object):
    def __init__(self, mod, titleTemplate):
        self.mod = mod
        self.inst = self.mod.Class(None)
        self.name = self.inst.name()
        self.appendExtraBlankLine = False
        self.titleTemplate = string.Template(titleTemplate)
        self.lines = []

    def appendLine(self, line, indent=0):
        line = line.strip()
        indent = '  ' * indent
        # this looked like a good idea, but it's actually breaking lines
        # in the middle of ReST elements:
        #lines = textwrap.wrap(line, 79,
        #                      initial_indent=indent,
        #                      subsequent_indent=indent)
        lines = [indent + line.replace('\n', '\n' + indent)]
        self.lines.extend(lines)
        if self.appendExtraBlankLine:
            self.lines.append('')

    def renderRST(self):
        self.appendExtraBlankLine = False
        self.appendLine('.. _plugin-%s:' % self.name)
        self.lines.append('')
        s = self.titleTemplate.substitute(name=self.name)
        self.appendLine(s)
        self.appendLine('=' * len(s))
        self.lines.append('')
        self.appendLine('Purpose')
        self.appendLine('-------')
        self.lines.append('')
        pdoc = getattr(self.mod, '__doc__',
                       'My author didn\'t give me a purpose.')
        self.appendLine(pdoc)
        self.lines.append('')
        cdoc = getattr(self.mod.Class, '__doc__', None)
        if cdoc is not None:
            self.appendLine('Usage')
            self.appendLine('-----')
            self.lines.append('')
            # We add spaces at the beginning in case the docstring does not
            # start with a newline (the default, unfortunately)
            self.appendLine(textwrap.dedent(" "* 1000 + cdoc).lstrip())
            self.lines.append('')
        commands = self.inst.listCommands()
        if len(commands):
            self.lines.append('.. _commands-%s:\n' % self.name)
            self.appendLine('Commands')
            self.appendLine('--------')
            self.lines.append('')
            for command in commands:
                log.debug('command: %s', command)
                self.lines.append('.. _command-%s-%s:\n' %
                    (self.name.lower(), command.replace(' ', '.')))
                line = '%s ' % command
                command = command.split()
                doc = self.inst.getCommandHelp(command)
                if doc:
                    doc = doc.replace('\x02', '')
                    (args, help) = doc.split(')', 1)
                    args = args.split('(', 1)[1]
                    args = args[len(' '.join(command)):].strip()
                    help = help.split('--', 1)[1].strip()
                    self.appendLine('``' + line + args + '``')
                    self.appendLine(help, 1)
                else:
                    self.appendLine('No help associated with this command')
                self.lines.append('')
        # now the config
        try:
            confs = conf.supybot.plugins.get(self.name)
            self.lines.append('.. _conf-%s:\n' % self.name)
            self.appendLine('Configuration')
            self.appendLine('-------------')
            self.lines.append('')
        except registry.NonExistentRegistryEntry:
            log.info('No configuration for plugin %s', plugin)
            self.appendLine('No configuration for this plugin')
        else:
            for confValues in self.genConfig(confs, 0):
                (name, kind, help, default, indent) = confValues
                self.appendLine('.. _conf-%s:' % name, indent - 1)
                self.lines.append('\n')
                self.appendLine('%s' % name, indent - 1)
                if default:
                    self.appendLine(
                        ('This config variable defaults to %s, %s')
                        % (default, kind), indent)
                else:
                    self.appendLine('This %s' % kind, indent)
                if help:
                    self.lines.append('')
                    self.appendLine(help, indent)
                self.lines.append('')
        return '\n'.join(self.lines) + '\n'

    def renderSTX(self):
        self.appendExtraBlankLine = True
        self.appendLine(self.titleTemplate.substitute(name=self.name))
        self.appendLine('Purpose', 1)
        pdoc = getattr(self.mod, '__doc__',
                       'My author didn\'t give me a purpose.')
        self.appendLine(pdoc, 2)
        cdoc = getattr(self.mod.Class, '__doc__', None)
        if cdoc is not None:
            self.appendLine('Usage', 1)
            self.appendLine(cdoc, 2)
        commands = self.inst.listCommands()
        if len(commands):
            self.appendLine('Commands', 1)
            for command in commands:
                log.debug('command: %s', command)
                line = '* %s ' % command
                command = command.split()
                doc = self.inst.getCommandHelp(command)
                if doc:
                    doc = doc.replace('\x02', '')
                    (args, help) = doc.split(')', 1)
                    args = args.split('(', 1)[1]
                    args = args[len(' '.join(command)):].strip()
                    help = help.split('--', 1)[1].strip()
                    self.appendLine(line + args, 2)
                    self.appendLine(help, 3)
                else:
                    self.appendLine('No help associated with this command', 3)
        # now the config
        try:
            confs = conf.supybot.plugins.get(self.name)
            self.appendLine('Configuration', 1)
        except registry.NonExistentRegistryEntry:
            log.info('No configuration for plugin %s', plugin)
            self.appendLine('No configuration for this plugin', 2)
        else:
            for confValues in self.genConfig(confs, 2):
                (name, kind, help, default, indent) = confValues
                self.appendLine('* %s' % name, indent - 1)
                if default:
                    self.appendLine(
                        ('This config variable defaults to %s, %s')
                        % (default, kind), indent)
                else:
                    self.appendLine('This %s' % kind, indent)
                self.appendLine(help, indent)
        return '\n'.join(self.lines) + '\n'

    def genConfig(self, item, origindent):
        confVars = item.getValues(getChildren=False, fullNames=False)
        if not confVars:
            return
        for (c, v) in confVars:
            name = v._name
            indent = origindent + 1
            if isinstance(v, registry.Value):
                try:
                    default = utils.str.dqrepr(str(v))
                    help = v.help()
                except registry.NonExistentRegistryEntry:
                    pass
                else:
                    cv = '' if v._channelValue else ' not'
                    nv = '' if v._networkValue else ' not'
                    kind = (
                        'is%s network-specific, and is%s channel-specific.'
                        % (nv, cv)
                    )
            else:
                help = ''
                default = ''
                kind = 'is a group of:'
            yield (name, kind, help, default, indent)
            for confValues in self.genConfig(v, indent):
                yield confValues

def genDoc(m, options):
    Plugin = PluginDoc(m, options.titleTemplate)
    outputFilename = string.Template(options.outputFilename).substitute(
        name=Plugin.name, format=options.format)
    path = os.path.join(options.outputDir, outputFilename)
    print('Generating documentation for %s to %s...' % (Plugin.name, path))
    try:
        fd = open(path, 'w')
    except EnvironmentError as e:
        error('Unable to open %s for writing.' % path)
    f = getattr(Plugin, 'render%s' % options.format.upper(), None)
    if f is None:
        fd.close()
        error('Unknown render format: `%s\'' % options.format)
    try:
        fd.write(f())
    finally:
        fd.close()

def main():
    import glob
    import os.path
    import optparse
    import supybot.plugin as plugin

    parser = optparse.OptionParser(usage='Usage: %prog [options] [plugins]',
                                   version='Supybot %s' % conf.version)
    parser.add_option('-c', '--clean', action='store_true', default=False,
                      dest='clean', help='Cleans the various data/conf/logs '
                      'directories after generating the docs.')
    parser.add_option('-o', '--output-dir', dest='outputDir', default='.',
                      help='Specifies the directory in which to write the '
                      'documentation for the plugin.')
    parser.add_option('--output-filename', dest='outputFilename',
                      default='$name.$format',
                      help='Specifies the path of individual output files. '
                      'If present in the value, "$name" will be substituted '
                      'with the plugin\'s name and "$format" with the value '
                      'if --format.')
    parser.add_option('-f', '--format', dest='format', choices=['rst', 'stx'],
                      default='rst', help='Specifies which output format to '
                      'use.')
    parser.add_option('--plugins-dir',
                      action='append', dest='pluginsDirs', default=[],
                      help='Looks in in the given directory for plugins and '
                      'generates documentation for all of them.')
    parser.add_option('--title-template',
                      default=TITLE_TEMPLATE, dest='titleTemplate',
                      help='Template string for the title of generated files')
    (options, args) = parser.parse_args()

    # This must go before checking for args, of course.
    for pluginDir in options.pluginsDirs:
        for name in glob.glob(os.path.join(pluginDir, '*')):
            if os.path.isdir(name):
                args.append(name)

    if not args:
        parser.print_help()
        sys.exit(-1)

    args = [s.rstrip('\\/') for s in args]
    pluginDirs = set([os.path.dirname(s) or '.' for s in args])
    conf.supybot.directories.plugins.setValue(list(pluginDirs))
    pluginNames = set([os.path.basename(s) for s in args])
    plugins = set([])
    for pluginName in pluginNames:
        if pluginName == '__pycache__':
            continue
        if pluginName.endswith('.py'):
            pluginName = pluginName[:-3]
        try:
            pluginModule = plugin.loadPluginModule(pluginName)
        except ImportError as e:
            s = 'Failed to load plugin %s: %s\n' \
                '(pluginDirs: %s)' % (pluginName, e,
                                      conf.supybot.directories.plugins())
            error(s)
        plugins.add(pluginModule)

    for Plugin in plugins:
        genDoc(Plugin, options)

    if options.clean:
        # We are about to remove the log dir; so trying to write anything
        # (such as "Shutdown initiated." and friends, from atexit callbacks)
        # would result in errors.
        log._handler.setLevel(logging.CRITICAL)
        shutil.rmtree(conf.supybot.directories.log())
        shutil.rmtree(conf.supybot.directories.conf())
        shutil.rmtree(conf.supybot.directories.data())


if __name__ == '__main__':
    main()

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=78:



###
# Copyright (c) 2003-2004, Jeremiah Fincher
# Copyright (c) 2009, James McCoy
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

"""
This is the main program to run Supybot.
"""

import re
import os
import sys
import atexit
import shutil
import signal

from io import StringIO  # Import this after version check since this will fail on Python 2

def _termHandler(signalNumber, stackFrame):
    raise SystemExit('Signal #%s.' % signalNumber)

signal.signal(signal.SIGTERM, _termHandler)

import time
import optparse
import textwrap

started = time.time()

import supybot
import supybot.utils as utils
import supybot.registry as registry
import supybot.questions as questions
import supybot.ircutils as ircutils
try:
    import supybot.i18n as i18n
except ImportError:
    sys.stderr.write("""Error:
    You are running a mix of Limnoria and stock Supybot code. Although you run
    one of Limnoria\'s executables, Python tries to load stock
    Supybot\'s library. To fix this issue, uninstall Supybot
    ("%s -m pip uninstall supybot" should do the job)
    and install Limnoria again.
    For your information, Supybot's libraries are installed here:
    %s\n""" %
    (sys.executable, '\n    '.join(supybot.__path__)))
    exit(-1)

from supybot.version import version

def run():
    import supybot.log as log

    log.info("Starting Limnoria %s on Python %s", version, sys.version)

    import supybot.conf as conf
    import supybot.world as world
    import supybot.drivers as drivers
    import supybot.ircmsgs as ircmsgs
    import supybot.schedule as schedule
    import supybot.httpserver as httpserver
    # We schedule this event rather than have it actually run because if there
    # is a failure between now and the time it takes the Owner plugin to load
    # all the various plugins, our registry file might be wiped.  That's bad.
    interrupted = False
    when = conf.supybot.upkeepInterval()
    schedule.addPeriodicEvent(world.upkeep, when, name='upkeep', now=False)
    world.startedAt = started
    while world.ircs:
        try:
            drivers.run()
        except KeyboardInterrupt:
            if interrupted:
                # Interrupted while waiting for queues to clear.  Let's clear
                # them ourselves.
                for irc in world.ircs:
                    irc._reallyDie()
                    continue
            else:
                interrupted = True
                log.info('Exiting due to Ctrl-C.  '
                         'If the bot doesn\'t exit within a few seconds, '
                         'feel free to press Ctrl-C again to make it exit '
                         'without flushing its message queues.')
                world.upkeep()
                for irc in world.ircs:
                    quitmsg = conf.supybot.plugins.Owner.quitMsg() or \
                              'Ctrl-C at console.'
                    # Because we're quitting from the console, none of the
                    # standard msg substitutions exist, and these will show as
                    # raw strings by default. Substitute them here with
                    # something meaningful instead.
                    env = dict((key, '<console>')
                               for key in ('who', 'nick', 'user', 'host'))
                    quitmsg = ircutils.standardSubstitute(irc, None, quitmsg,
                                                          env=env)
                    irc.queueMsg(ircmsgs.quit(quitmsg))
                    irc.die()
        except SystemExit as e:
            s = str(e)
            if s:
                log.info('Exiting due to %s', s)
            break
        except:
            try: # Ok, now we're *REALLY* paranoid!
                log.exception('Exception raised out of drivers.run:')
            except Exception as e:
                print('Exception raised in log.exception.  This is *really*')
                print('bad.  Hopefully it won\'t happen again, but tell us')
                print('about it anyway, this is a significant problem.')
                print('Anyway, here\'s the exception: %s' % \
                      utils.gen.exnToString(e))
            except:
                print('Oh, this really sucks.  Not only did log.exception')
                print('raise an exception, but freaking-a, it was a string')
                print('exception.  People who raise string exceptions should')
                print('die a slow, painful death.')
    httpserver.stopServer()
    now = time.time()
    seconds = now - world.startedAt
    log.info('Total uptime: %s.', utils.gen.timeElapsed(seconds))
    (user, system, _, _, _) = os.times()
    log.info('Total CPU time taken: %.2f seconds.', user+system)
    log.info('No more Irc objects, exiting.')

def main():
    parser = optparse.OptionParser(usage='Usage: %prog [options] configFile',
                                   version='Limnoria %s running on Python %s' %
                                   (version, sys.version))
    parser.add_option('-P', '--profile', action='store_true', dest='profile',
                      help='enables profiling')
    parser.add_option('-n', '--nick', action='store',
                      dest='nick', default='',
                      help='nick the bot should use')
    parser.add_option('-u', '--user', action='store',
                      dest='user', default='',
                      help='full username the bot should use')
    parser.add_option('-i', '--ident', action='store',
                      dest='ident', default='',
                      help='ident the bot should use')
    parser.add_option('-d', '--daemon', action='store_true',
                      dest='daemon',
                      help='Determines whether the bot will daemonize.  '
                           'This is a no-op on non-POSIX systems.')
    parser.add_option('', '--allow-default-owner', action='store_true',
                      dest='allowDefaultOwner',
                      help='Determines whether the bot will allow its '
                           'defaultCapabilities not to include "-owner", thus '
                           'giving all users the owner capability by default. '
                           'This is a security risk since it allows anyone to run '
                           'any command on your bot, so we advise not to use this.')
    parser.add_option('', '--allow-root', action='store_true',
                      dest='allowRoot',
                      help='Determines whether the bot will be allowed to run '
                           'as root. This should not be used except in special '
                           'circumstances, such as running inside a containerized '
                           'environment.')
    parser.add_option('', '--debug', action='store_true', dest='debug',
                      help='Determines whether some extra debugging stuff '
                      'will be logged in this script.')
    parser.add_option('', '--disable-multiprocessing', action='store_true',
                      dest='disableMultiprocessing',
                      help='Disables multiprocessing stuff. May lead to '
                      'vulnerabilities.')

    (options, args) = parser.parse_args()

    if os.name == 'posix':
        if (os.getuid() == 0 or os.geteuid() == 0) and not options.allowRoot:
            sys.stderr.write('Running as root is not supported by default (see --allow-root).')
            sys.stderr.write(os.linesep)
            sys.exit(-1)

    if len(args) > 1:
        parser.error("""Only one configuration file should be specified.""")
    elif not args:
        parser.error(utils.str.normalizeWhitespace("""It seems you've given me
        no configuration file.  If you do have a configuration file, be sure to
        specify the filename.  If you don't have a configuration file, read
        docs/GETTING_STARTED and follow the instructions."""))
    else:
        registryFilename = args.pop()
        try:
            # The registry *MUST* be opened before importing log or conf.
            registry.open_registry(registryFilename)
            shutil.copyfile(registryFilename, registryFilename + '.bak')
        except registry.InvalidRegistryFile as e:
            s = '%s in %s.  Please fix this error and start supybot again.' % \
                (e, registryFilename)
            s = textwrap.fill(s)
            sys.stderr.write(s)
            sys.stderr.write(os.linesep)
            raise
            sys.exit(-1)
        except EnvironmentError as e:
            sys.stderr.write(str(e))
            sys.stderr.write(os.linesep)
            sys.exit(-1)

    i18n.getLocaleFromRegistryCache()

    try:
        import supybot.log as log
    except supybot.registry.InvalidRegistryValue as e:
        # This is raised here because supybot.log imports supybot.conf.
        name = e.value._name
        errmsg = textwrap.fill('%s: %s' % (name, e),
                               width=78, subsequent_indent=' '*len(name))
        sys.stderr.write(errmsg)
        sys.stderr.write(os.linesep)
        sys.stderr.write('Please fix this error in your configuration file '
                         'and restart your bot.')
        sys.stderr.write(os.linesep)
        sys.exit(-1)
    import supybot.conf as conf
    import supybot.world as world
    i18n.import_conf()
    world.starting = True

    def closeRegistry():
        # We only print if world.dying so we don't see these messages during
        # upkeep.
        logger = log.debug
        if world.dying:
            logger = log.info
        logger('Writing registry file to %s', registryFilename)
        registry.close(conf.supybot, registryFilename)
        logger('Finished writing registry file.')
    world.flushers.append(closeRegistry)
    world.registryFilename = registryFilename

    nick = options.nick or conf.supybot.nick()
    user = options.user or conf.supybot.user()
    ident = options.ident or conf.supybot.ident()

    networks = conf.supybot.networks()
    if not networks:
        questions.output("""No networks defined.  Perhaps you should re-run the
        wizard?""", fd=sys.stderr)
        # XXX We should turn off logging here for a prettier presentation.
        sys.exit(-1)

    if os.name == 'posix' and options.daemon:
        def fork():
            child = os.fork()
            if child != 0:
                if options.debug:
                    print('Parent exiting, child PID: %s' % child)
                # We must us os._exit instead of sys.exit so atexit handlers
                # don't run.  They shouldn't be dangerous, but they're ugly.
                os._exit(0)
        fork()
        os.setsid()
        # What the heck does this do?  I wonder if it breaks anything...
        # ...It did.  I don't know why, but it seems largely useless.  It seems
        # to me reasonable that we should respect the user's umask.
        #os.umask(0)
        # Let's not do this for now (at least until I can make sure it works):
        # Actually, let's never do this -- we'll always have files open in the
        # bot directories, so they won't be able to be unmounted anyway.
        # os.chdir('/')
        fork()
        # Since this is the indicator that no writing should be done to stdout,
        # we'll set it to True before closing stdout et alii.
        conf.daemonized = True
        # Closing stdin shouldn't cause problems.  We'll let it raise an
        # exception if it does.
        sys.stdin.close()
        # Closing these two might cause problems; we log writes to them as
        # level WARNING on upkeep.
        sys.stdout.close()
        sys.stderr.close()
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        # We have to be really methodical here.
        os.close(0)
        os.close(1)
        os.close(2)
        fd = os.open('/dev/null', os.O_RDWR)
        os.dup2(fd, 0)
        os.dup2(fd, 1)
        os.dup2(fd, 2)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)
        log.info('Completed daemonization.  Current PID: %s', os.getpid())

    # Stop setting our own umask.  See comment above.
    #os.umask(077)

    # Let's write the PID file.  This has to go after daemonization, obviously.
    pidFile = conf.supybot.pidFile()
    if pidFile:
        try:
            fd = open(pidFile, 'w')
            pid = os.getpid()
            fd.write('%s%s' % (pid, os.linesep))
            fd.close()
            def removePidFile():
                try:
                    os.remove(pidFile)
                except EnvironmentError as e:
                    log.error('Could not remove pid file: %s', e)
            atexit.register(removePidFile)
        except EnvironmentError as e:
            log.critical('Error opening/writing pid file %s: %s', pidFile, e)
            sys.exit(-1)

    conf.allowDefaultOwner = options.allowDefaultOwner
    world.disableMultiprocessing = options.disableMultiprocessing

    if not os.path.exists(conf.supybot.directories.log()):
        os.mkdir(conf.supybot.directories.log())
    if not os.path.exists(conf.supybot.directories.conf()):
        os.mkdir(conf.supybot.directories.conf())
    if not os.path.exists(conf.supybot.directories.data()):
        os.mkdir(conf.supybot.directories.data())
    if not os.path.exists(conf.supybot.directories.data.tmp()):
        os.mkdir(conf.supybot.directories.tmp())

    userdataFilename = os.path.join(conf.supybot.directories.conf(),
                                    'userdata.conf')
    # Let's open this now since we've got our directories setup.
    if not os.path.exists(userdataFilename):
        fd = open(userdataFilename, 'w')
        fd.write('\n')
        fd.close()
    registry.open_registry(userdataFilename)

    import supybot.irclib as irclib
    import supybot.ircmsgs as ircmsgs
    import supybot.drivers as drivers
    import supybot.callbacks as callbacks
    import supybot.plugins.Owner as Owner

    # This may take some resources, and it does not need to run while booting, so
    # we import it as late as possible (but before plugins are loaded).
    import supybot.httpserver as httpserver

    owner = Owner.Class()

    if options.profile:
        import cProfile
        world.profiling = True
        cProfile.runctx('run()',
                        globals=globals(), locals={**locals(), "run": run},
                        filename='%s-%i.prof' % (nick, time.time()))
    else:
        run()


# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



#!/usr/bin/env python3

###
# Copyright (c) 2002-2004, Jeremiah Fincher
# Copyright (c) 2010-2021, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###



import supybot

from supybot.questions import *

import os
import sys
import optparse

def main():
    import supybot.log as log
    import supybot.conf as conf
    conf.supybot.log.stdout.setValue(False)
    parser = optparse.OptionParser(usage='Usage: %prog [options] <users.conf>',
                                   version='supybot %s' % conf.version)
    parser.add_option('-u', '--username', action='store', default='',
                      dest='name',
                      help='username for the user.')
    parser.add_option('-p', '--password', action='store', default='',
                      dest='password',
                      help='password for the user.')
    parser.add_option('-c', '--capability', action='append',
                      dest='capabilities', metavar='CAPABILITY',
                      help='capability the user should have; '
                           'this option may be given multiple times.')
    (options, args) = parser.parse_args()
    if len(args) != 1:
        parser.error('Specify the users.conf file you\'d like to use.  '
                     'Be sure *not* to specify your registry file, generated '
                     'by supybot-wizard.  This is not the file you want.  '
                     'Instead, take a look in your conf directory (usually '
                     'named "conf") and take a gander at the file '
                     '"users.conf".  That\'s the one you want.')

    filename = os.path.abspath(args[0])
    conf.supybot.directories.log.setValue('/')
    conf.supybot.directories.conf.setValue('/')
    conf.supybot.directories.data.setValue('/')
    conf.supybot.directories.plugins.setValue(['/'])
    conf.supybot.databases.users.filename.setValue(filename)
    import supybot.ircdb as ircdb

    if not options.name:
        name = ''
        while not name:
            name = something('What is the user\'s name?')
            try:
                # Check to see if the user is already in the database.
                _ = ircdb.users.getUser(name)
                # Uh oh.  That user already exists;
                # otherwise we'd have KeyError'ed.
                output('That user already exists.  Try another name.')
                name = ''
            except KeyError:
                # Good.  No such user exists.  We'll pass.
                pass
    else:
        try:
            # Same as above. We exit here instead.
            _ = ircdb.users.getUser(options.name)
            output('That user already exists.  Try another name.')
            sys.exit(-1)
        except KeyError:
            name = options.name

    if not options.password:
        password = getpass('What is %s\'s password? ' % name)
    else:
        password = options.password

    if not options.capabilities:
        capabilities = []
        prompt = 'Would you like to give %s a capability?' % name
        while yn(prompt):
            capabilities.append(anything('What capability?'))
            prompt = 'Would you like to give %s another capability?' % name
    else:
        capabilities = options.capabilities

    user = ircdb.users.newUser()
    user.name = name
    user.setPassword(password)
    for capability in capabilities:
        user.addCapability(capability)
    ircdb.users.setUser(user)
    ircdb.users.flush()
    #os.system('cat %s' % filename) # Was this here just for debugging?
    ircdb.users.close()
    print('User %s added.' % name)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass

# vim:set shiftwidth=4 softtabstop=4 expandtab textwidth=79:



###
# Copyright (c) 2022, Valentin Lorentz
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions, and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions, and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the author of this software nor the name of
#     contributors to this software may be used to endorse or promote products
#     derived from this software without specific prior written consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

###

import re
import datetime

# Credits for the regexp and function: https://stackoverflow.com/a/2765366/539465
_XSD_DURATION_RE = re.compile(
    r"(?P<sign>-?)P"
    r"(?:(?P<years>\d+)Y)?"
    r"(?:(?P<months>\d+)M)?"
    r"(?:(?P<days>\d+)D)?"
    r"(?:T(?:(?P<hours>\d+)H)?(?:(?P<minutes>\d+)M)?(?:(?P<seconds>\d+)S)?)?"
)


def parse_xsd_duration(s):
    """Parses this format to a timedelta:
    https://www.w3.org/TR/xmlschema11-2/#duration"""
    # Fetch the match groups with default value of 0 (not None)
    duration = _XSD_DURATION_RE.match(s).groupdict(0)

    # Create the timedelta object from extracted groups
    delta = datetime.timedelta(
        days=int(duration["days"])
        + (int(duration["months"]) * 30)
        + (int(duration["years"]) * 365),
        hours=int(duration["hours"]),
        minutes=int(duration["minutes"]),
        seconds=int(duration["seconds"]),
    )

    if duration["sign"] == "-":
        delta *= -1

    return delta
